
/* Chrysalide - Outil d'analyse de fichiers binaires
 * maccess.c - accès à la mémorie à partir d'un registre et d'un décallage
 *
 * Copyright (C) 2014-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "maccess.h"


#include "../../../operand-int.h"
#include "../../../sharing/manager.h"
#include "../../../../common/sort.h"



/* --------------------- MANIPULATION D'OPERANDES INDIVIDUELLES --------------------- */


/* Définition d'un opérande offrant un accès à la mémoire depuis une base (instance) */
struct _GArmV7MAccessOperand
{
    GArchOperand parent;                    /* Instance parente            */

    GArchOperand *base;                     /* Base de l'accès en mémoire  */
    GArchOperand *offset;                   /* Décallage pour l'adresse    */
    GArchOperand *shift;                    /* Décallage pour le décallage */
    bool not_post_indexed;                  /* Positio du décallage        */
    bool write_back;                        /* Mise à jour de la base      */

};


/* Définition d'un opérande offrant un accès à la mémoire depuis une base (classe) */
struct _GArmV7MAccessOperandClass
{
    GArchOperandClass parent;               /* Classe parente              */

};


/* Initialise la classe des accès à la mémoire chez ARM. */
static void g_armv7_maccess_operand_class_init(GArmV7MAccessOperandClass *);

/* Initialise une instance d'accès à la mémoire chez ARM. */
static void g_armv7_maccess_operand_init(GArmV7MAccessOperand *);

/* Supprime toutes les références externes. */
static void g_armv7_maccess_operand_dispose(GArmV7MAccessOperand *);

/* Procède à la libération totale de la mémoire. */
static void g_armv7_maccess_operand_finalize(GArmV7MAccessOperand *);

/* Initialise un nouvel objet partagé avec des informations. */
static bool g_armv7_maccess_operand_apply_template(GArmV7MAccessOperand *, const GArmV7MAccessOperand *);

/* Réalise une copie minimale d'un contenu partagé. */
static void g_armv7_maccess_operand_define_template(const GArmV7MAccessOperand *, GArmV7MAccessOperand *);

/* Compare un opérande avec un autre. */
static int g_armv7_maccess_operand_compare(const GArmV7MAccessOperand *, const GArmV7MAccessOperand *);

/* Traduit un opérande en version humainement lisible. */
static void g_armv7_maccess_operand_print(const GArmV7MAccessOperand *, GBufferLine *, AsmSyntax);



/* -------------------------- PARTAGES DE CONTENUS UNIQUES -------------------------- */


/* Gestionnaire des partages d'instances */
static GShareManager *_armv7_maccess_operand_manager = NULL;


/* Fournit le gestionnaire de partages attribué à un type. */
static GShareManager *get_armv7_maccess_operand_share_manager(void);



/* ---------------------------------------------------------------------------------- */
/*                       MANIPULATION D'OPERANDES INDIVIDUELLES                       */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini par la GLib pour un accès à la mémoire depuis une base. */
G_DEFINE_TYPE(GArmV7MAccessOperand, g_armv7_maccess_operand, G_TYPE_ARCH_OPERAND);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des accès à la mémoire chez ARM.        *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_maccess_operand_class_init(GArmV7MAccessOperandClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GArchOperandClass *operand;             /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);
    operand = G_ARCH_OPERAND_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_armv7_maccess_operand_dispose;
    object->finalize = (GObjectFinalizeFunc)g_armv7_maccess_operand_finalize;

    operand->get_manager = (get_operand_manager_fc)get_armv7_maccess_operand_share_manager;

    operand->apply_template = (apply_operand_template_fc)g_armv7_maccess_operand_apply_template;
    operand->define_template = (define_operand_template_fc)g_armv7_maccess_operand_define_template;
    operand->free_template = (free_operand_template_fc)NULL;

    operand->compare = (operand_compare_fc)g_armv7_maccess_operand_compare;
    operand->print = (operand_print_fc)g_armv7_maccess_operand_print;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance à initialiser.                            *
*                                                                             *
*  Description : Initialise une instance d'accès à la mémoire chez ARM.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_maccess_operand_init(GArmV7MAccessOperand *operand)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_maccess_operand_dispose(GArmV7MAccessOperand *operand)
{
    g_object_unref(G_OBJECT(operand->base));

    if (operand->offset != NULL)
        g_object_unref(G_OBJECT(operand->offset));

    if (operand->shift != NULL)
        g_object_unref(G_OBJECT(operand->shift));

    G_OBJECT_CLASS(g_armv7_maccess_operand_parent_class)->dispose(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = instance d'objet GLib à traiter.                   *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_maccess_operand_finalize(GArmV7MAccessOperand *operand)
{
    G_OBJECT_CLASS(g_armv7_maccess_operand_parent_class)->finalize(G_OBJECT(operand));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand  = objet partagé à initialiser.                      *
*                template = information à utiliser pour la mise en place.     *
*                                                                             *
*  Description : Initialise un nouvel objet partagé avec des informations.    *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_armv7_maccess_operand_apply_template(GArmV7MAccessOperand *operand, const GArmV7MAccessOperand *template)
{
    g_armv7_maccess_operand_define_template(template, operand);

    g_object_ref(G_OBJECT(operand->base));

    if (operand->offset != NULL)
        g_object_ref(G_OBJECT(operand->offset));

    if (operand->shift != NULL)
        g_object_ref(G_OBJECT(operand->shift));

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand  = objet partagé à consulter.                        *
*                template = informations à retrouver intégralement.           *
*                                                                             *
*  Description : Réalise une copie minimale d'un contenu partagé.             *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_maccess_operand_define_template(const GArmV7MAccessOperand *operand, GArmV7MAccessOperand *template)
{
    template->base = operand->base;
    template->offset = operand->offset;
    template->shift = operand->shift;

    template->not_post_indexed = operand->not_post_indexed;
    template->write_back = operand->write_back;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : a = premier opérande à consulter.                            *
*                b = second opérande à consulter.                             *
*                                                                             *
*  Description : Compare un opérande avec un autre.                           *
*                                                                             *
*  Retour      : Bilan de la comparaison.                                     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int g_armv7_maccess_operand_compare(const GArmV7MAccessOperand *a, const GArmV7MAccessOperand *b)
{
    int result;                             /* Bilan à faire remonter      */

    result = g_arch_operand_compare(a->base, b->base);
    if (result != 0) goto gamoc_done;

    result = sort_pointer(a->offset, b->offset, (__compar_fn_t)g_arch_operand_compare);
    if (result != 0) goto gamoc_done;

    result = sort_pointer(a->shift, b->shift, (__compar_fn_t)g_arch_operand_compare);
    if (result != 0) goto gamoc_done;

    result = sort_boolean(a->not_post_indexed, b->not_post_indexed);
    if (result != 0) goto gamoc_done;

    result = sort_boolean(a->write_back, b->write_back);

 gamoc_done:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à traiter.                                *
*                line    = ligne tampon où imprimer l'opérande donné.         *
*                syntax  = type de représentation demandée.                   *
*                                                                             *
*  Description : Traduit un opérande en version humainement lisible.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_armv7_maccess_operand_print(const GArmV7MAccessOperand *operand, GBufferLine *line, AsmSyntax syntax)
{
    g_buffer_line_append_text(line, BLC_ASSEMBLY, "[", 1, RTT_HOOK, NULL);

    g_arch_operand_print(operand->base, line, syntax);

    if (!operand->not_post_indexed)
        g_buffer_line_append_text(line, BLC_ASSEMBLY, "]", 1, RTT_HOOK, NULL);

    if (operand->offset != NULL)
    {
        g_buffer_line_append_text(line, BLC_ASSEMBLY, ",", 1, RTT_PUNCT, NULL);
        g_buffer_line_append_text(line, BLC_ASSEMBLY, " ", 1, RTT_RAW, NULL);

        g_arch_operand_print(operand->offset, line, syntax);

    }

    if (operand->shift != NULL)
    {
        g_buffer_line_append_text(line, BLC_ASSEMBLY, ",", 1, RTT_PUNCT, NULL);
        g_buffer_line_append_text(line, BLC_ASSEMBLY, " ", 1, RTT_RAW, NULL);

        g_arch_operand_print(operand->shift, line, syntax);

    }

    if (operand->not_post_indexed)
        g_buffer_line_append_text(line, BLC_ASSEMBLY, "]", 1, RTT_HOOK, NULL);

    if (operand->write_back)
        g_buffer_line_append_text(line, BLC_ASSEMBLY, "!", 1, RTT_PUNCT, NULL);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : base    = représente le registre de la base d'accès.         *
*                offset  = détermine le décallage entre l'adresse et la base. *
*                shift   = opération de décallage pour jouer sur le décallage.*
*                indexed = précise la forme donnée au décallage à appliquer.  *
*                wback   = indique une mise à jour de la base après usage.    *
*                                                                             *
*  Description : Crée un accès à la mémoire depuis une base et un décallage.  *
*                                                                             *
*  Retour      : Opérande mis en place.                                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchOperand *g_armv7_maccess_operand_new(GArchOperand *base, GArchOperand *offset, GArchOperand *shift, bool indexed, bool wback)
{
    GSharedInstance *result;                /* Structure à retourner       */
    GArmV7MAccessOperand template;          /* Transport d'informations    */

    template.base = base;
    template.offset = offset;
    template.shift = shift;

    template.not_post_indexed = indexed;
    template.write_back = wback;

    result = g_share_manager_build(_armv7_maccess_operand_manager, (GSharedInstance *)&template);

    return G_ARCH_OPERAND(result);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à consulter.                              *
*                                                                             *
*  Description : Founit la base d'un accès à la mémoire.                      *
*                                                                             *
*  Retour      : Opérande en place.                                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchOperand *g_armv7_maccess_operand_get_base(const GArmV7MAccessOperand *operand)
{
    return operand->base;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à consulter.                              *
*                                                                             *
*  Description : Founit le décallage d'un accès à la mémoire depuis la base.  *
*                                                                             *
*  Retour      : Opérande en place.                                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchOperand *g_armv7_maccess_operand_get_offset(const GArmV7MAccessOperand *operand)
{
    return operand->offset;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à consulter.                              *
*                                                                             *
*  Description : Founit le décallage d'un décallage pour un accès mémoire.    *
*                                                                             *
*  Retour      : Opérande en place.                                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GArchOperand *g_armv7_maccess_operand_get_shift(const GArmV7MAccessOperand *operand)
{
    return operand->shift;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à consulter.                              *
*                                                                             *
*  Description : Indique si le décallage est post-indexé.                     *
*                                                                             *
*  Retour      : Statut des opérations menées.                                *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_armv7_maccess_operand_is_post_indexed(const GArmV7MAccessOperand *operand)
{
    return !operand->not_post_indexed;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operand = opérande à consulter.                              *
*                                                                             *
*  Description : Indique si la base est mise à jour après usage.              *
*                                                                             *
*  Retour      : Statut des opérations menées.                                *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_armv7_maccess_operand_has_to_write_back(const GArmV7MAccessOperand *operand)
{
    return operand->write_back;

}



/* ---------------------------------------------------------------------------------- */
/*                            PARTAGES DE CONTENUS UNIQUES                            */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit le gestionnaire de partages attribué à un type.      *
*                                                                             *
*  Retour      : Gestionnaire de partages en place.                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GShareManager *get_armv7_maccess_operand_share_manager(void)
{
    return _armv7_maccess_operand_manager;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Met en place les mécanismes de partage des accès mémmoire.   *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool init_armv7_maccess_operand_sharing(void)
{
    _armv7_maccess_operand_manager = g_share_manager_new(G_TYPE_ARMV7_MACCESS_OPERAND);

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Imprime des statistiques quant aux partages dans l'archi.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/
#ifdef DEBUG_DUMP_STATS
void dump_armv7_maccess_operand_share_stats(void)
{
    g_share_manager_dump_stats(_armv7_maccess_operand_manager);

}
#endif


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Supprime les mécanismes de partage des accès mémoire.        *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void exit_armv7_maccess_operand_sharing(void)
{
    g_object_unref(G_OBJECT(_armv7_maccess_operand_manager));

}
